%% e-TeX V2.0;2
% Note: the first line of this file must match the corresponding line
% in "etex.src" and "etexdefs.lib", otherwise fallback will be used.

% $Id: language.us.def 18737 2010-06-04 17:09:02Z karl $
% language.us.def (and the start of language.def)
% - initial hyphenation patterns.

% DO NOT EDIT THIS FILE (language.def)! It is generated by tlmgr.
% See language.dat (or language.us) for more information.

% It is explicitly permitted to augment this file by adding additional
% \addlanguage ... commands AFTER the first (\addlanguage {USenglish})
% which must remain as the first language added to ensure total document
% portability for pre-\language documents.  The last line of the file
% MUST be left as \uselanguage {USenglish} for the sane reason.

% Additional languages, patterns, exceptions and left- and right-hyphen
% minima may be added here, and an example is given for German.

% The five parameters for \addlanguage are:
% language, patterns file, exceptions file, left- and right-hyphen minima

%%% Next section is      E X A M P L E   O N L Y
%%% \addlanguage {German}{ghyph31}{}{2}{2}   %%% further such lines may be used;
%%% Previous section is  E X A M P L E   O N L Y

\addlanguage {USenglish}{hyphen}{}{2}{3} %%% This MUST be the first non-comment
                                         %%% line of the file

\uselanguage {USenglish} %%% This MUST be the last line of the file.
